#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <stdio.h>    
#include <stdlib.h>
#include "control.h"
#include "model.h"
#include "indiv.h"
#include "diploid.h"
#include "haploid.h"
#include "fpmath.h"
#include "fp.h" 

using namespace std;

CtrlParam::CtrlParam(void)
{
	;
}

CtrlParam::~CtrlParam(void)
{
	;
} 

void CtrlParam::OnHelp(void)
{
	PrintHeader(); 
	cout << " " << endl;
	cout << " FILE I/O RELATED OPTIONS." << endl;
	cout << " -g <filename>    " << " repeatable for multiple input files. must pair with -p" << endl;
	cout << " -p <0/1/10/11/12/...>  " << " must pair with -g, <1> pairing genotypes are cohort" << endl; 
	cout << "    <0> pairing genotypes are panel but no labelling, <10/11/12/...> pairing genotype are labelled panels." << endl;
	cout << " -pos <file>      " << " repeatable for multiple input files" << endl; 
    cout << " -o <prefix>      " << " prefix of all output files, random seeds will be used by default" << endl;  
	cout << " -FILE <filename>    " << " specify a phenotype file." << endl;
	cout << " -DOC <filename>    " << " specify a file containing covariates to be controlled for." << endl;
	cout << " -nph <num>  " << " number of phenotypes to be analyzed jointly."  << endl; 
	cout << " " << endl; 
	cout << " EM RELATED OPTIONS." << endl;
	cout << " -e <num>   " << " specify number of EM runs, default is 1." << endl; 
	cout << " -w <num>   " << " specify steps of EM run, default 30, using fast linear approximation." << endl; 
	cout << " -s <num>   " << " specify steps of EM run, default 0, using quadratic algorithm." << endl; 
	cout << " -C <num>         " << " specify number of upper clusters, default 2" << endl; 
	cout << " -c <num>         " << " specify number of lower clusters, default 10" << endl; 
	cout << " -mixgen <num>         " << " specify number of generations the admixture occured" << endl; 
	cout << " -R <num>         " << " specify random seed, system time by default" << endl; 
	cout << " -sem <num>      " << " to save EM results to prefix.em.txt, if 0 save after warm up EM, if 1 save after joint EM." << endl; 
	cout << " -rem <file>      " << " to read EM results" << endl; 
//	cout << " -minit <1/2>      " << " 2: initalize at all markers, 1: (default) only initialize a random marker" << endl; 
	cout << " " << endl; 
//	cout << " --ps1       " << " output marginal p(s|data) for diploid, mean local ancestry."  << endl; 
//	cout << " --ps2       " << " output joint p(s1,s2|data) for diploid."  << endl; 
//	cout << " --pk1       " << " output marginal p(k|data) for diploid individual."  << endl; 
//	cout << " --pk2       " << " linear algorithm for model fitting, viterbi path to phase and compute p(k|haplotype)"  << endl; 
	cout << " " << endl;

	cout << " " << endl;
	cout << " OTHER OPTIONS." << endl;
	cout << " --ver          " << "print version and citation" << endl;
	cout << " --h(help)         " << "print this help" << endl;
//	cout << " -gene <file>     " << " to read gene file that specify regions of interests" << endl; 
//	cout << " -gf(GF) <num>    " << " pair with -gene to specify gene flanking region in kb" << endl; 
	cout << " -exclude-maf <num> " << endl; 
	cout << "          exclude SNPs whose maf < num, default 0.01" << endl; 
	cout << " -exclude-miss <num> " << endl; 
	cout << "          exclude SNPs whose missing rate > num, default 1" << endl; 
	cout << " --exclude-miss1 " << endl; 
	cout << "          exclude SNPs that are missing in one file" << endl; 
//	cout << " --silence        " << "no terminal output" <<  endl; 
	cout << " " << endl; 
	cout << " " << endl; 
}

void CtrlParam::PrintHeader(void)
{
	cout << endl; 
	cout << " hapQTL version 1.00, visit http://haplotype.org for possible update." << endl;
	cout << " Developed by Yongtao Guan ytguan@gmail.com, Baylor Collge of Medicine. All Rights Reserved." << endl; 
	cout << " Last updated on 9 May 2016." << endl; 
	cout << endl; 
}

void CtrlParam::BatchRun(int argc, char ** argv)
{
	pMD = new class ModelnData;
                             
	int rem = 0; 
	int sem = 0; 
	int ps1 = 0; 
	int pk1 = 0; 
	int admix = 0; 
	int file = 0; 

//	int bf = 0; 
	int mph = 0; 

//	int radmix = 0; 
//	int ps2 = 0; 
//	int pk2 = 0; 
//	int lhs = 0; 
//	int error = 0; 
//	int sel = 0; 
//	int qc = 0; 

	int npath = 2; 
	int silence = 0; 
	int yes_weg = 0; 
	int yes_wmg = 0; 
	int yes_wbg = 0;   
	int yes_wgd = 0; 
	int weg_letter = 1; 
	int all_wmg = 1; 
	int all_wbg = 0;   //best guess default only write cohort snps; 
	int all_wgd = 1; 

	string fnEM; 
	string fnAdmix; 
	string note; 
    string fnmap; 

	for(int i = 1; i < argc; i++) 
	{   
		string str;  
		string opt; 
		if(argv[i][0] != '-') 
			continue;
		
		str.assign(argv[i]);
		
		if(str.compare("--silence") == 0) {
			silence = 1; 
			pMD->Setsilence(1);
		} 
		else if (str.compare("-minit") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			pMD->set_minit(atoi(argv[i+1])); 
		}
		else if (str.compare("--verbose") == 0) {
			pMD->Setverbose();
		}
		else if (str.compare("-exclude-maf") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			pMD->Setexcludemaf(atof(argv[i+1]));
		}
		else if (str.compare("-exclude-miss") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			pMD->Setexcludemiss(atof(argv[i+1]));
		}
		else if (str.compare("--exclude-miss1") == 0) {
		   	pMD->Setexcludemiss1();
		}
//		else if (str.compare("--exclude-nopos") == 0) {
//			pMD->Setexcludenopos();
//		}
//		else if (str.compare("-allele-coding") == 0) {
//			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
//			pMD->Setallelecoding(atoi(argv[i+1]));
//		}
		else if (str.compare("-note") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			note.assign(argv[i+1]);
		}
		else if (str.compare("-FILE") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetfnFILE(ts);
			file = 1; 
		}
		else if (str.compare("-DOC") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetfnDOC(ts);
		}
//		else if (str.compare("-radmix") == 0) {
//			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
//			string ts(argv[i+1]);
//			fnAdmix.assign(ts);
//			radmix = 1; 
//		}
		else if (str.compare("-rph") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetfnPhenotype(ts);
		}
		else if (str.compare("-rcov") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetfnCovariate(ts);
		}
		else if (str.compare("-g") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetvGin(ts);
		}
		else if (str.compare("-p") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetvPin(ts);
		}
		else if (str.compare("-pos") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetvPos(ts);
		}
		else if (str.compare("-o") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			string ts(argv[i+1]);
			pMD->SetfnOutput(ts);
		}
		else if (str.compare("-R") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			pMD->SetrandSeed(atoi(argv[i+1]));
		}
		else if (str.compare("-e") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			pMD->SetnEMRuns(atoi(argv[i+1]));
		}
		else if (str.compare("-num") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong augument after option." << endl;
				exit(0);
			}
			pMD->Setnum(atof(argv[i+1]));
		}
		else if (str.compare("-nph") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong augument after option." << endl;
				exit(0);
			}
			pMD->set_nph(atoi(argv[i+1]));
		}
		else if (str.compare("-s") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			pMD->SetnMaxSteps(atoi(argv[i+1]));
		}
		else if (str.compare("-w") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			pMD->SetnWarmSteps(atoi(argv[i+1]));
		}
		else if (str.compare("-c") == 0) {
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				if(!isdigit(argv[i+1][0]))
				{
					cout << "wrong argument after option." << endl; 
					exit(0); 
				}
				pMD->SetnK(atoi(argv[i+1]));
		}
		else if (str.compare("-C") == 0) {
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				if(!isdigit(argv[i+1][0]))
				{
					cout << "wrong argument after option." << endl; 
					exit(0); 
				}
				pMD->SetnS(atoi(argv[i+1]));
		}
		else if (str.compare("-mg") == 0 || str.compare("-mixgen") == 0) {
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				if(!isdigit(argv[i+1][0]))
				{
					cout << "wrong argument after option." << endl; 
					exit(0); 
				}
				pMD->set_mixgen(atof(argv[i+1]));
		}
		else if (str.compare("-npath") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			npath = atoi(argv[i+1]);
			pMD->set_npath(npath);
		}
		else if (str.compare("-rem") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			fnEM.assign(argv[i+1]);
			rem = 1;
		}
		else if (str.compare("-sem") == 0) {
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			sem += atoi(argv[i+1]);
			sem = 1;
		}
		else if (str.compare("-morgan") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			pMD->set_morgan(atof(argv[i+1]));
		}
//		else if (str.compare("-rm") == 0) {
//			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
//			fnmap.assign(argv[i+1]);
//		}
		else if (str.compare("-mu") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			if(!isdigit(argv[i+1][0]))
			{
				cout << "wrong argument after option." << endl; 
				exit(0); 
			}
			pMD->set_mu(atof(argv[i+1]));
		}
//		else if (str.compare("-rr") == 0) {
//			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
//			if(!isdigit(argv[i+1][0]))
//			{
//				cout << "wrong argument after option." << endl; 
//				exit(0); 
//			}
//			pMD->set_upper_switches(atof(argv[i+1]));
//		}
//		else if (str.compare("-rrt") == 0) {
//			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
//			if(!isdigit(argv[i+1][0]))
//			{
//				cout << "wrong argument after option." << endl; 
//				exit(0); 
//			}
//			pMD->set_rrt(atoi(argv[i+1]));
//		}
		else if (str.compare("-h") == 0 || str.compare("-help") == 0) {
				OnHelp();
				exit(0); 
		}
		else if (str.compare("-v") == 0 || str.compare("-ver") == 0) {
				PrintHeader(); 
				exit(0); 
		}
		else if (str.compare("--disable-hap") == 0) {
				pMD->set_disable_hap(); 
		}
		else if (str.compare("-weg") == 0) {
				yes_weg = 1; 
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				weg_letter = atoi(argv[i+1]); 
		}
		else if (str.compare("-wmg") == 0) {
				yes_wmg = 1; 
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				all_wmg = atoi(argv[i+1]); 
		}
		else if (str.compare("-wbg") == 0) {
				yes_wbg = 1; 
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				all_wbg = atoi(argv[i+1]); 
		}
		else if (str.compare("-wgd") == 0) {
				yes_wgd = 1; 
				if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
				all_wgd = atoi(argv[i+1]); 
		}
		else if (str.compare("--ps1") == 0) 
			ps1 = 1; 
//		else if (str.compare("--ps2") == 0) 
//			ps2 = 1; 
		else if (str.compare("--pk1") == 0) 
			pk1 = 1; 
//		else if (str.compare("--pk2") == 0) 
//			pk2 = 1; 
		else if (str.compare("--admix") == 0) 
			admix = 1; 
		else if (str.compare("--mph") == 0) 
			mph = 1; 
//		else if (str.compare("--lhs") == 0) 
//			lhs += 1; 
//		else if (str.compare("--qc") == 0) 
//			qc = 1; 
//		else if (str.compare("--error") == 0) 
//			error = 1; 
//		else if (str.compare("--sel") == 0) 
//			sel = 1; 
		else if (str.compare("-setc") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			pMD->set_ctrl_index(atoi(argv[i+1])); 
		}
		else if (str.compare("-setx") == 0) {
			if(argv[i+1] == NULL || argv[i+1][0] == '-') continue;
			pMD->set_design_index(atoi(argv[i+1])); 
		}
		else {
			fprintf(stderr,"Bad option %s\n", argv[i]);
			OnHelp();
			exit(0); 
		}              
	}
    

	streambuf * cout_strbuf(cout.rdbuf()); 
	//save current cout into a streambuf; 
	ostringstream output; 
	if(silence == 1) 
		cout.rdbuf(output.rdbuf()); 
	//redirect cout to a streambuf output; 


	int randseed = pMD->GetrandSeed();  // default returns -1;
	if (randseed <= 0)
		randseed = (unsigned) time(NULL);
	gsl_rng_set(gsl_r, randseed);


	string prefix(pMD->GetfnOutput()); 
	if(prefix.compare("\0") == 0) {
		char tmp[100];
		memset(tmp, '\0', 100);
		sprintf(tmp, "%d", randseed); 
		prefix.assign(tmp); 
		pMD->SetfnOutput(prefix); 
	}
	//if no random seed provided, use system clock;    
	//if no prefix provided, use randseed; 
	
	pMD->open_log(); 
	(pMD->fplog) << "## " << note << endl;
	(pMD->fplog) << "## COMMAND: "; 
	for (int i = 0; i < argc; i++)
		(pMD->fplog) << argv[i] << " "; 
	(pMD->fplog) << endl; 
	(pMD->fplog) << "## randseed = " << randseed << endl; 
	cout << "## randseed = " << randseed << endl; 

	///////////////////////////////
//	if(lhs > 1)  
//	{
//		cout << "combining LHS: -num specify number of individuals, -g input lhs files, -o output lhs+ prefix " << endl; 
//		pMD->combine_lhs(1); 
//		exit(0); 
//	}
	//////////////////////////////
//	{
//		pMD->play_ground(); 
//		exit(0); 
//	}
	

    pMD->read_bimbam_genotype(0, -1); 
	///////////////////////////////
//	{
//		pMD->simulate_admixed(); 
//		exit(0); 
//	}
	//////////////////////////////

	time_t sec_beg, sec_end; 
	sec_beg = time(NULL); 
	
	{
		if(fnmap.size() > 0)
			pMD->read_rs2map(fnmap); 

		if(rem == 0)
			pMD->InitModelParam(); 
		else 
			pMD->read_em(fnEM); 
//		if(radmix == 1) 
//			pMD->read_admix(fnAdmix); 

		if(yes_weg == 1) pMD->write_genotype(0, weg_letter); 

		pMD->EM_joint(1);  
		pMD->EM_joint(2);  
		pMD->write_em(1); 
		pMD->write_admix(); 

		if(yes_wmg == 1) 
		{
			pMD->compute_mgt(npath); 
			pMD->write_genotype(1, all_wmg); 
		}
		if(yes_wbg == 1) pMD->write_genotype(2, all_wbg); 
		if(yes_wgd == 1) pMD->write_genotype(3, all_wgd); 
				  
//		if(lhs == 1) 
//			pMD->compute_lhs(npath); 
		if(pk1 == 1) 
			pMD->compute_pk(1, npath); 
		if(ps1 == 1) 
			pMD->compute_ps(1, npath); 
		if(file == 0) {
			cout << "no phenotype file" << endl; 
			exit(0); 
		}
		if(mph == 1) 
			pMD->hpass_matrixnormal(npath); 
		else if (admix == 1) 
			pMD->hpass_admix(npath); 
		else
			pMD->hpass_pk(npath); 
//		if(error == 1) 
//			pMD->detect_error(npath); 
		//this is for genotyping error detection, not for debugging. 
	}

	sec_end = time(NULL); 
	(pMD->fplog) << "## EM seconds used = " << sec_end - sec_beg << endl; 
	
	pMD->fplog << "## Program generated the following files in the output directory." << endl; 
	pMD->fplog << "## " << prefix << ".log.txt" << endl; 
	pMD->fplog << "## " << prefix << ".snpinfo.txt" << endl; 
	pMD->fplog << "## " << prefix << ".em.txt" << endl; 
	pMD->fplog << "## " << prefix << ".admix.txt" << endl; 
	if(yes_weg == 1) pMD->fplog << "## "<< prefix << ".exact.genotype.txt" << endl;
	if(yes_wmg == 1) pMD->fplog << "## "<< prefix << ".mean.genotype.txt" << endl;
	if(yes_wbg == 1) pMD->fplog << "## "<< prefix << ".best.guess.genotype.txt" << endl; 
	if(yes_wgd == 1) pMD->fplog << "## "<< prefix << ".genotype.distribution.txt" << endl; 
	pMD->fplog << "## " << prefix << ".bf.txt" << endl; 
	
	(pMD->fplog) << "## random seed = " << randseed << endl;
	pMD->close_log();
	cout << "-finished. for details see log: "<<  prefix << ".log.txt" << endl; 

	if(silence == 1)
		cout.rdbuf(cout_strbuf); 
	//restore the original cout; 
//	delete pMD; 
}

